// ====================== /elements/local PacketPacer.hh ===============================
// 
// Written by: Alexander Jacobsen
// Element Description: Element which increase the pacing between packets leaving
//						the element. The time a packet leaves the element is stored
//						and a metric is used in combination with the leave time of
//						the packet in order to set the leave time of the next packet.
// Last Updated: 06.06.2011
//
// ========================================================================================

#ifndef CLICK_PacketPacer_HH
#define CLICK_PacketPacer_HH
#include <click/element.hh>
#include <click/straccum.hh>
#include <click/standard/storage.hh>
#include <click/timer.hh>

//All click declarations are enclosed within a macro pair, CLICK_DECLS and CLICKENDDECLS.
CLICK_DECLS

	class PacketPacer : public Element { public:

			PacketPacer();
			~PacketPacer();
			
			//Returns class name.
			const char *class_name() const { return "PacketPacer"; }
			
			//Sets the elements ports. This element has one input and one output port.
			const char *port_count() const { return "1/1"; }

			//Sets the elements port behavior. This element has push ports.
			const char *processing() const { return  "h/lh"; }

			//initialize() not defined. Default used.
			int initialize(ErrorHandler*);

			//methods defining the ports of the element.
			void push(int port, Packet *p);
			Packet * pull(int);

			//Add a handler which makes it possible to read the nr of packets.
			void add_handlers();

			int averageTimeBetweenAcks() const { return _averageTimeBetweenAcks; }

			//Function to put the packets into the queue.
			inline bool enqueue(Packet*);

			//Function which sets the leave time for each packet.
			inline void	setLeaveTimeForPacket(int);

			//Function which calculates the leave time for each packet.
			inline Timestamp calculateLeaveTimeForPacket(int);

			static String read_handler(Element*, void*);

			void cleanup(CleanupStage);

			int calculateAverageBetweenACKs();

			void PacketPacer::storeDestIPsToManipulate();

			inline bool PacketPacer::checkIfPacketBelongsToFlowOfInterest();

			void PacketPacer::run_timer(Timer *timer);

			inline Packet* PacketPacer::decreaseQueue(Storage*, Packet** );

			bool _checkForPacketsInQueue(Storage*);

			inline int PacketPacer::getKpMetric();


			private:

			int _totalPacketCounter;
			int _averagePacketCounter;
			short _new_capacity;

			StringAccum sa;
			
			//Dynamically allocate memory for the packet leave times.
			Timestamp *packetLeaveTime;
			Timestamp *latestAckLeaveTimeForFlow;
			String *listOfDestinationIPsToManipulate;
			String *listOfSourceIPsToManipulate;
			int *kpMetrics;
			
			//Timestamps.
			Timestamp pullTimeStamp;
			Timestamp leaveTime;
			Timestamp timeCheck;
			Timestamp newTimeStamp;
			Timestamp oldTimeStamp;
			Timestamp totalTimeStampBetweenAcks;
			Timestamp tempTimeBetweenThisAndPreviousPack;

			//Offsets
			uint8_t _annotationOffset;

			int _averageTimeBetweenAcks;
			int _drops;
			int _highwater_length;
			int _leaveTime;
			bool _increaseAckDelay;

			String _sessionInfo;
			int _nrOfIps;
			String _packSrcIpAddress, _packDestinationAddress;

			String _logg;
			int _loggInterval;
			Timer _loggTimer;
			int _leaveTimeIncrease;

			//Storage controllers used to set the capacity, control where packets are
			//stored in the queue. Need one for each queue.
			Storage _primaryStorage;
			Storage _secondaryStorage;

			Packet* volatile * _primaryQueue;
			Packet* volatile * _secondaryQueue;

};

CLICK_ENDDECLS
#endif


