/* ====================== elements/local/PacketDelayElement.cc ================
 *
 *
 * Element Description: Element which rewrites the window field in the TCP header.
 * 						A handler (serviceSession) is used to get information used
 * 						to identify flows and calculate a new window size given in
 * 						bytes. This window size adjustment is used to control the
 * 						bit rate of different TCP flows.
 *
 *						Parameters retrieved from the handler are:
 *						1. Source IP
 *						2. Destination IP
 *						3. Round-trip time (ms)
 *						4. Target bit rate (bps)
 *
 * NOTE: 				This element assume that window scaling is not in use.
 *
 * Written by: Alexander Jacobsen
 *
 * Last Updated: 12.05.2011
 *
 * ============================================================================
 */

#include<click/config.h>
#include <clicknet/tcp.h>
#include <clicknet/ip.h>
#include <click/ipaddress.hh>
#include <click/confparse.hh>
#include<click/error.hh>
#include<click/string.hh>
#include<click/timestamp.hh>
#include <click/straccum.hh>

#include "TcpWindowFieldReWriter.hh"

/*
 * All click declarations are enclosed within a macro pair, CLICK_DECLS and
 * CLICKENDDECLS.
 */
CLICK_DECLS

/*
 *	Constructor which set all startup values and initiate arrays needed
 *	to store information during the run-time of the element.
 */
TcpWindowFieldReWriter::TcpWindowFieldReWriter()
{
	packetCounter = 0;
	nrOfServiceSessions = 0;

	serviceSessions = "129.241.197.82:128.205.32.24:1:1000000,";
	//serviceSessions = "";

	//Arrays which store information retrieved from the handler (serviceSessions).
	listOfDestinationIPsToManipulate = new String[1000];
	listOfSourceIPsToManipulate = new String[1000];
	listOfRTTs = new int[1000];
	listOfTargetBitRates = new int[1000];
}

/*
 *  Destructor
 */
TcpWindowFieldReWriter::~TcpWindowFieldReWriter(){ }

/*
 * Initializer is the last place to check the element for errors before it is
 * started. Not used in this element.
 *
 * Returns 0 if no errors where found.
 *
 */
int TcpWindowFieldReWriter::initialize(ErrorHandler *errh)
{
	return 0;
}

/*
 * Handlers which makes it possible to change the values of variables run-time.
 */
void TcpWindowFieldReWriter::add_handlers()
{
	add_data_handlers("serviceSessions", Handler::OP_READ | Handler::OP_WRITE, &serviceSessions);
}

/*
 * simple_action defines the ports of the element as agnostic. This means that it is the context
 * of the configuration file that determines if the element ports are push or pull.
 */
Packet * TcpWindowFieldReWriter::simple_action(Packet *p){

	//If packet counter reaches its maximum value it should be reset.
	if(packetCounter == 2147483646)
	{
		packetCounter=0;
	}

	//Counter used for packet identification in the debug text.
	packetCounter++;

	bool _packBelongsToAServiceFlow = false;
	int _newWindowSize = 0;
	int _indexOfServiceFlow = 0;
	unsigned _currentTcpWindowSize;

	retriveServiceSessionParamsFromHandler();

	// Pointer to the TCP header
	click_tcp *tcpHeader = p->tcp_header();

	// Pointer to the IP header
	click_ip *ipHeader = p->ip_header();

	ipSrcAddress = IPAddress(ipHeader->ip_src);
	ipDestAddress = IPAddress(ipHeader->ip_dst);

	unsigned plen = ntohs(ipHeader->ip_len) - (ipHeader->ip_hl << 2);
	unsigned packetCheckSum;

	_currentTcpWindowSize = ntohs(tcpHeader->th_win);

	_debugText << "******************** Packet nr: " << packetCounter << " **************************" << "\n";
	_debugText << "IP src addr: " << ipSrcAddress << " --- IP dest addr: " << ipDestAddress << "\n";

	_packBelongsToAServiceFlow = checkIfPacketBelongsToServiceFlow(ipSrcAddress, ipDestAddress);

	//Check if packet belongs to a service flow of interest. True: get the index of the parameters (ip, rtt, bitrate, etc) of this
	//service flow stored in arrays.
	if(_packBelongsToAServiceFlow)
	{
		_debugText << "### Packet is RECOGNIZED as part of a service flow ###" << "\n";

		_indexOfServiceFlow = getIndexOfServiceFlowInInformationArrays(ipSrcAddress,ipDestAddress);

		_newWindowSize = calculateWindowSize(listOfTargetBitRates[_indexOfServiceFlow], listOfRTTs[_indexOfServiceFlow] );

		tcpHeader->th_win = ntohs(_newWindowSize);

		//Recalculate and set checksum.
		tcpHeader->th_sum = 0;
		packetCheckSum = click_in_cksum((unsigned char *)tcpHeader, plen);
		tcpHeader->th_sum = click_in_cksum_pseudohdr(packetCheckSum, ipHeader, plen);
	}
	else
	{
		_debugText << "### Packet is NOT RECOGNIZED as part of a service flow ###" << "\n";
		_newWindowSize = _currentTcpWindowSize;
	}

	_debugText << "Current window size = " << _currentTcpWindowSize << "\n";
	_debugText << "Calculated window size = " << _newWindowSize << "\n";

	//_debugText << "ACK sequence nr : " << ntohs(tcpHeader->th_ack) << "\n";
	//_debugText << "ACK nr : " << (tcpHeader->th_ack) << "\n";

	//Write general debug text to the click chatter.
	click_chatter(_debugText.c_str());
	_debugText.clear();

	return p;
}

/*
 * Function which clears the queues from memory.
 */
void TcpWindowFieldReWriter::cleanup(CleanupStage){}

/*
 * Function which retrieves all parameters from the service session
 * handler and stores them in corresponding arrays.
 */
void TcpWindowFieldReWriter::retriveServiceSessionParamsFromHandler(){

	int _stringSize = serviceSessions.length();
	int _commaSeparatorPosIterator = 0;
	int _colonSeparatorPos = 0;
	int _startIndexOfDestIP = 0;
	int _lengthOfDestIP = 0;
	int _lengthOfRTT = 0;
	int _startIndexOfRTT = 0;
	int _previousColonSeparatorPos = 0;
	int _startIndexOfTargetBitRate = 0;
	int _lengthOfTargetBitRate = 0;
	String _rest = serviceSessions;
	String _sourceIP = "";
	String _destIP = "";
	String _tempKpTargetRate = "";
	String _tempKpRtt = "";
	long targetRate = 0;

	//Search for comma in string from handler. Returns -1 if no comma is found.
	int _commaSeparatorPos = _rest.find_left(',' , 0);

	//Double parser not working as it should....
	//double rtt;

	int rtt;

	//When the handler is read then the number of service session are reset.
	nrOfServiceSessions = 0;

	//Iterating the string of service session parameters. The string is on the following form:
	//sourceIP:destIP:RTT:targetBitRate,sourceIP:destIP:RTT:targetBitRate -->
	//The iteration runs as long as a comma is found in the string.
	while(_commaSeparatorPos > 0) {

		//Search for char in string. Returns -1 if the character is not found.
		_colonSeparatorPos = _rest.find_left(':' , 0);
		_commaSeparatorPos = _rest.find_left(',' , 0);

		//Do something if the handler contain service sessions.
		if(_commaSeparatorPos >= 0){

			//Extract source IP address from handler string.
			_sourceIP = _rest.substring(0,_colonSeparatorPos);

			_previousColonSeparatorPos = _colonSeparatorPos;

			//Update colon separator with the position of the next colon separator.
			//(Here: after the destination IP address.)
			_colonSeparatorPos = _rest.find_left(':' , _previousColonSeparatorPos+1);

			_startIndexOfDestIP = _previousColonSeparatorPos+1;
			_lengthOfDestIP = _colonSeparatorPos-_startIndexOfDestIP;

			//Extract destination IP from handler string.
			_destIP = _rest.substring(_startIndexOfDestIP, _lengthOfDestIP);

			_previousColonSeparatorPos = _colonSeparatorPos;

			//Find next colon in string.
			_colonSeparatorPos = _rest.find_left(':' , _previousColonSeparatorPos+1);

			//Update colon separator with the position of the next colon separator.
			//(Here: after the RTT.)
			_startIndexOfRTT = _previousColonSeparatorPos+1;
			_lengthOfRTT = _colonSeparatorPos-_startIndexOfRTT;

			_tempKpRtt = _rest.substring(_startIndexOfRTT, _lengthOfRTT );

			_previousColonSeparatorPos = _colonSeparatorPos;

			_startIndexOfTargetBitRate = _previousColonSeparatorPos+1;
			_lengthOfTargetBitRate = _commaSeparatorPos-_startIndexOfTargetBitRate;

			_tempKpTargetRate = _rest.substring(_startIndexOfTargetBitRate, _lengthOfTargetBitRate );


			//Store all parameters in corresponding arrays.
			listOfSourceIPsToManipulate[nrOfServiceSessions] = _sourceIP;
			listOfDestinationIPsToManipulate[nrOfServiceSessions] = _destIP;

			//Convert RTT from string to double.
			cp_integer(_tempKpRtt, 0, &rtt);
			listOfRTTs[nrOfServiceSessions] = rtt;

			//Convert target rate from string to integer.
			cp_integer(_tempKpTargetRate, 0, &targetRate);
			listOfTargetBitRates[nrOfServiceSessions] = targetRate;

			_commaSeparatorPosIterator += _commaSeparatorPos+1;

			//The remaining service flow parameters are stored in _rest to be retrieved
			//in the next round of this loop.
			_rest = serviceSessions.substring(_commaSeparatorPosIterator,_stringSize);

			_stringSize = _rest.length();

			nrOfServiceSessions++;
		}
	}

}
/*
 * Function which checks if the source and destination IP is recognized.
 * If it is, then a new receive window should be calculated.
 */
inline bool TcpWindowFieldReWriter::checkIfPacketBelongsToServiceFlow(String sourceIP, String destIp)
{
	bool _isFlowRecognized = false;

	//Loop through all service sessions if they exist.
	if (nrOfServiceSessions != 0)
	{
		for (int i = 0; i < nrOfServiceSessions; i++)
		{
			if ( listOfSourceIPsToManipulate[i].compare(sourceIP) == 0 && listOfDestinationIPsToManipulate[i].compare(destIp) == 0 )
			{
				_isFlowRecognized = true;
			}
		}
	}
	return _isFlowRecognized;
}
/*
 *  Function which takes the source and destination IP of a service session and finds
 *  their index in the arrays where all parameters are stored.
 *
 */
inline int TcpWindowFieldReWriter::getIndexOfServiceFlowInInformationArrays(String sourceIP, String destIp)
{
	int serviceFlowIndex = -1;

	//Loop through all destination IP's if they exist
	if (nrOfServiceSessions != 0)
	{
		for (int i = 0; i < nrOfServiceSessions; i++)
		{
			if ( listOfSourceIPsToManipulate[i].compare(sourceIP) == 0 && listOfDestinationIPsToManipulate[i].compare(destIp) == 0)
			{
				serviceFlowIndex = i;
			}
		}
	}
	return serviceFlowIndex;
}

/*
 * A function which calculates new window size given the target bit rate (bps) and RTT (ms).
 * Returns the new window size (bytes).
 */
inline long TcpWindowFieldReWriter::calculateWindowSize(long newBitRate, int RTT)
{
	long _calculatedWindowSize = 0;
	_calculatedWindowSize = (newBitRate*RTT)/(8*1000);
	return _calculatedWindowSize;
}

CLICK_ENDDECLS
ELEMENT_REQUIRES(linuxmodule)
//Needed for Clicks compilation process.
EXPORT_ELEMENT(TcpWindowFieldReWriter)

