/* ====================== elements/local/PacketDelayElement.hh ================
 *
 *
 * Element Description: Element which rewrites the window field in the TCP header.
 * 						A handler (serviceSession) is used to get information used
 * 						to identify flows and calculate a new window size given in
 * 						bytes. This window size adjustment is used to control the
 * 						bit rate of different TCP flows.
 *
 *						Parameters retrieved from the handler are:
 *						1. Source IP
 *						2. Destination IP
 *						3. Round-trip time (ms)
 *						4. Target bit rate (bps)
 *
 * NOTE: 				This element assume that window scaling is not in use.
 *
 * Written by: Alexander Jacobsen
 *
 * Last Updated: 12.05.2011
 *
 * ============================================================================
 */

#ifndef CLICK_TcpWindowFieldReWriter_HH
#define CLICK_TcpWindowFieldReWriter_HH

#include <click/element.hh>
#include <click/straccum.hh>

//All click declarations are enclosed within a macro pair, CLICK_DECLS and CLICKENDDECLS.
CLICK_DECLS

// The class is defined in the header file (C++)
class TcpWindowFieldReWriter : public Element { public:

	TcpWindowFieldReWriter();
	~TcpWindowFieldReWriter();

	//Returns class name.
	const char *class_name() const { return "TcpWindowFieldReWriter"; }

	//Sets the elements ports. This element has one input and one output port.
	const char *port_count() const { return "1/1"; }

	//Sets the elements port behavior. This element has push ports.
	const char *processing() const { return "h/h"; }

	//initialize() not defined. Default used.
	int initialize(ErrorHandler*);

	Packet * TcpWindowFieldReWriter::simple_action(Packet *p);

	//Add a handler which makes it possible to read the nr of packets.
	void add_handlers();

	//Default used.
	void TcpWindowFieldReWriter::cleanup(CleanupStage);

	void TcpWindowFieldReWriter::storeDestIPsToManipulate();
	void TcpWindowFieldReWriter::retriveServiceSessionParamsFromHandler();
	inline long TcpWindowFieldReWriter::calculateWindowSize(long, int);
	inline bool TcpWindowFieldReWriter::checkIfPacketBelongsToServiceFlow(String sourceIP, String DestIp);
	inline int TcpWindowFieldReWriter::getIndexOfServiceFlowInInformationArrays(String sourceIP, String DestIp);

private:

	long int packetCounter;
	int _newTcpWindowSize;
	int nrOfServiceSessions;

	//Dynamically allocate memory for the packet leave times.
	String *listOfDestinationIPsToManipulate;
	String *listOfSourceIPsToManipulate;
	int *listOfRTTs;
	long *listOfTargetBitRates;

	String serviceSessions;
	String ipSrcAddress, ipDestAddress;
	StringAccum _debugText;

};

CLICK_ENDDECLS
#endif


