package no.ntnu.item.router;

import static no.ntnu.item.router.Constants.CLICK_CHANGE_INTERVAL;
import static no.ntnu.item.router.Constants.CLICK_HANDLER_PATH;

import java.io.FileWriter;
import java.util.ArrayList;
import java.util.List;

/**
 * This class will have an overview of the {@link ServiceSession} objects<br>
 * that have a target bit rate set. It will, based on this information<br>
 * change parameters in the click router to facilitate the wanted effect.<br>
 * It will continuously communicate with click and decide when to do what.
 * 
 * @author Kristian Haugene
 * 
 */
public class ActionReasoner extends Thread {

	/** Holds the list of {@link ServiceSession} objects to monitor for */
	private List<ServiceSession> markedSessions;

	/** list of {@link ServiceSession} to store updates before processing */
	private List<ServiceSession> updatedSessions;

	/** Indicates if the list has been changed since last configuration */
	private boolean listHasChanged;

	/** Current string that is written to the Click handler file */
	private List<SessionHandlerInfo> sessionHandlerInfoList;

	/**
	 * Constructor. Start the thread
	 */
	public ActionReasoner() {
		this.markedSessions = new ArrayList<ServiceSession>();
		this.updatedSessions = new ArrayList<ServiceSession>();
		this.listHasChanged = false;
		this.sessionHandlerInfoList = new ArrayList<SessionHandlerInfo>();

		this.start();
	}

	/**
	 * The thread main method. Loop to check for changes and calculate action.
	 */
	public void run() {

		// Infinate loop
		while (true) {

			// Check each second
			try {
				Thread.sleep(1000);
			} catch (Exception e) {
			}
			
			// If new lists, update, clear temp and change flag
			if (listHasChanged) {
				doListHasChangedRoutine();
			}

			// Check each session
			for (ServiceSession markedSession : markedSessions) {

				/*
				 * Before checking if targetRate is met, check if the user
				 * preferred target rate has changed. If it has we need to reset
				 * clickTargetBitRate and let it calculate again
				 */
				if (isTargetVideoBitrateChanged(markedSession)) {

					// Reset it
					markedSession.setClickTargetVideoBitRate(0);

					// Update the oldTargetVideoBitRate value to the current
					// value
					markedSession.setOldTargetVideoBitRate(markedSession
							.getTargetVideoBitrate());
				}

				if (targetRateIsMet(markedSession)) {
					// TODO REMOVE DEBUG INFO
					Integer targetRate = markedSession.getTargetVideoBitrate();
					Integer latestRate = markedSession.getLatestMeasurement()
							.getVideoDownloadBitrate();

					System.out.println("OK Target rate: " + targetRate
							+ ", DL rate: " + latestRate);
				} else {
					// TODO REMOVE DEBUG INFO
					Integer targetRate = markedSession.getTargetVideoBitrate();
					Integer latestRate = markedSession.getLatestMeasurement()
							.getVideoDownloadBitrate();

					System.out.println("NO! Target rate: " + targetRate
							+ ", DL rate: " + latestRate);

					if (targetRateIsLowerThanCurrentRate(markedSession)) {
						lowerSessionRate(markedSession);
					} else {
						increaseSessionRate(markedSession);
					}

				}
			}

			// Do Click handler update if necessary
			calculateClickHandlerChange();

		}

	}

	/**
	 * Update the {@link ActionReasoner} list of sessions.
	 * 
	 * @param newList
	 *            the updated list
	 */
	public void updateListOfMarkedSessions(List<ServiceSession> newList) {

		/*
		 * We write the new list to a temporary location. When the process
		 * notice the listHasChanged flag it will updates its reference. This
		 * way we avoid manipulating a list while the thread is working
		 */
		updatedSessions = newList;
		listHasChanged = true;

	}

	/**
	 * This method will increase the click target rate for a given session.
	 * 
	 * @param markedSession
	 *            to increase click rate for
	 */
	private void increaseSessionRate(ServiceSession markedSession) {

		// If click is not currently slowing it, nothing to do(in this version)
		// Hint: We could iterate other sessions, set their rate lower. Or
		// rewrite click conf.
		if (markedSession.getClickTargetVideoBitRate() == 0) {
			return;
		}

		/*
		 * We have set a click target value, and it has decreased the bandwidth
		 * beyond the wanted result. We need to increment it.
		 */

		incrementClickTargetRate(markedSession);

	}

	/**
	 * This method will increment click target rate for a session. It will be
	 * incremented given that there is at least 10 seconds since last change to
	 * the rate. If not enough time has passed, do nothing
	 * 
	 * @param session
	 *            to increment click target rate for
	 */
	private void incrementClickTargetRate(ServiceSession session) {

		// Get user target rate, just for setting it back later to reset timer
		Integer userTargetRate = session.getTargetVideoBitrate();

		// Calculate time in ms since last change
		Long msSinceLastChange = System.currentTimeMillis()
				- session.getTargetRateSetTime();

		// Convert to a integer number of seconds
		int secondsSinceLastChange = (int) (msSinceLastChange / 1000);

		// Check if the necessary time has passed to make a new change
		if (secondsSinceLastChange > CLICK_CHANGE_INTERVAL) {

			// It has, increment the rate
			Integer newRate = session.getClickTargetVideoBitRate() + 20000;
			session.setClickTargetVideoBitRate(newRate);

			// We have made an effort to reach target. Refresh target value.
			session.setTargetVideoBitrate(userTargetRate);
		}

	}

	/**
	 * This method will tell click to lower the rate for this session
	 * 
	 * @param session
	 *            the session to slow down
	 */
	private void lowerSessionRate(ServiceSession session) {

		if (session.getClickTargetVideoBitRate() == 0) {
			/*
			 * This session have not been slowed yet. Set an initial click
			 * target rate.
			 */
			setFirstLoweredClickTargetRate(session);

		} else {
			/*
			 * We have already tried to change the rate. But it was not enough.
			 * Decrement...
			 */
			decrementClickTargetRate(session);

		}

	}

	/**
	 * This method will decrement click target rate for a session. It will be
	 * decremented given that there is at least 10 seconds since last change to
	 * the rate. If not enough time has passed, do nothing
	 * 
	 * @param session
	 *            to decrement click target rate for
	 */
	private void decrementClickTargetRate(ServiceSession session) {

		// Get user target rate, just for setting it back later to reset timer
		Integer userTargetRate = session.getTargetVideoBitrate();

		// Calculate time in ms since last change
		Long msSinceLastChange = System.currentTimeMillis()
				- session.getTargetRateSetTime();

		// Convert to a integer number of seconds
		int secondsSinceLastChange = (int) (msSinceLastChange / 1000);

		// Check if the necessary time has passed to make a new change
		if (secondsSinceLastChange > CLICK_CHANGE_INTERVAL) {

			// It has, decrement the rate
			Integer newRate = session.getClickTargetVideoBitRate() - 20000;
			session.setClickTargetVideoBitRate(newRate);

			// We have made an effort to reach target. Refresh target value.
			session.setTargetVideoBitrate(userTargetRate);
		}
	}

	/**
	 * This method will set an initial click target value for a session. We have
	 * a distinct click target value so that we can vary this to obtain the user
	 * target value without actually changing the user target value and show
	 * this in the GUI. The initial click target is determined to be in between
	 * the user target and the next available bit rate. That is, if the user
	 * wants a quality level 800, and the next quality level is 1000 we will set
	 * 900. This is to give room for the 800 stream to function well, but not
	 * enough bandwidth to run the 1000 quality level.
	 * 
	 * @param session
	 *            the session to set initial click target value for
	 */
	private void setFirstLoweredClickTargetRate(ServiceSession session) {

		// Get available rates
		List<Integer> rates = session.getLatestMeasurement()
				.getAvaliableVideoBitrates();

		// Get the user preferred rate
		Integer targetRate = session.getTargetVideoBitrate();

		// Find the rate that is higher than the user preffered
		Integer oneRateAboveTarget = 0;
		for (int i = 0; i < rates.size(); i++) {
			if (targetRate.intValue() == rates.get(i).intValue()) {
				oneRateAboveTarget = rates.get(i + 1);
			}
		}

		// Calculate a value in between
		Integer inBetween = (targetRate + oneRateAboveTarget) / 2;

		// Set it
		session.setClickTargetVideoBitRate(inBetween);

		// We have made an effort to reach target. Refresh target value.
		session.setTargetVideoBitrate(targetRate);

	}

	/**
	 * Checks if a {@link ServiceSession} is downloading at its target rate
	 * 
	 * @param session
	 *            session to check for
	 * @return returns true if target rate is equal to download rate OR target
	 *         rate is not set. False otherwise
	 */
	private Boolean targetRateIsMet(ServiceSession session) {

		Integer targetRate = session.getTargetVideoBitrate();
		Integer latestRate = session.getLatestMeasurement()
				.getVideoDownloadBitrate();

		// If target rate is equal to latest rate, true. False otherwise
		return targetRate.intValue() == latestRate.intValue();
	}

	/**
	 * Checks if the target rate is lower than the actual rate, ie if we need to
	 * inhibit it.
	 * 
	 * @param session
	 *            session to check for
	 * @return returns true if target rate is less than actual, false if equal
	 *         or greater
	 */
	private Boolean targetRateIsLowerThanCurrentRate(ServiceSession session) {

		Integer targetRate = session.getTargetVideoBitrate();
		Integer latestRate = session.getLatestMeasurement()
				.getVideoDownloadBitrate();

		// If target rate is equal to latest rate, true. False otherwise
		return targetRate.intValue() < latestRate.intValue();

	}

	/**
	 * This method will check all sessions. Based on this info it will calculate
	 * the wanted parameter to pass to the Click handler. For each time this
	 * method is called it will do this calculation. If the calculated handler
	 * parameter varies from earlier values, the handler will be rewritten. <br>
	 * <br>
	 * The Click handler expects clientIp, sourceIp, RTT and targetRate. Values
	 * are separated by ':' and sets of values are separated by ','
	 */
	private void calculateClickHandlerChange() {

		// Get a list of SessionHandlerInfo based on the current data
		List<SessionHandlerInfo> newList = getNewListOfSessionHandlerInfo();

		// Check if the new list is equal to the last list(with margins on ping)
		Boolean listsAreEqual = areListsEqual(newList);

		if (listsAreEqual) {
			// do nothing
			return;
		}

		// lists are not equal, we should exchange them and rewrite handler
		this.sessionHandlerInfoList = newList;
		String newHandlerString = "";

		for (SessionHandlerInfo sessionHandlerInfo : sessionHandlerInfoList) {
			newHandlerString += sessionHandlerInfo.getClientAddress() + ":";
			newHandlerString += sessionHandlerInfo.getSourceAddress() + ":";

			// TODO DETTA MÅ RYDDES

			 newHandlerString += sessionHandlerInfo.getPing().toString() +
			 ":";

			// <NY>
			Double ping = sessionHandlerInfo.getPing();
			if (ping < 1) {
				newHandlerString += "1:";
			} else {
				newHandlerString += ping.toString() + ":";
			}
			// </NY>

			newHandlerString += sessionHandlerInfo.getClickTargetRate() + ",";
		}

		// TODO Remove debug info
		System.out.println("NEW HANDLER: " + newHandlerString);
		writeFile(CLICK_HANDLER_PATH, newHandlerString);

	}

	/**
	 * Will compare the given list with this.sessionHandlerInfoList<br>
	 * Will determine if the lists are approximately equal.<br>
	 * The reason we say approximately equal is that we allow some margins on
	 * ping. We want the ping to have changed more than a certain threshold
	 * before we rewrite the Click handler. <br>
	 * <br>
	 * The lists will have to be of equal length, containing the same objects,
	 * with the same target click rate. The ping values should also be
	 * relatively close. If they are equal, except from a small change in ping,
	 * we return true.
	 * 
	 * @param newList
	 *            the list to compare to this.sessionHandlerInfoList
	 * @return true if they are equal or have a small difference in ping values,
	 *         false otherwise
	 */
	private Boolean areListsEqual(List<SessionHandlerInfo> newList) {

		// If size is different, they are not equal
		if (newList.size() != this.sessionHandlerInfoList.size()) {
			return false;
		}

		// Assume that the entries are equal in-order, if not we rewrite
		for (int i = 0; i < newList.size(); i++) {

			Boolean sameClient = newList.get(i).getClientAddress()
					.equals(sessionHandlerInfoList.get(i).getClientAddress());

			Boolean sameSource = newList.get(i).getSourceAddress()
					.equals(sessionHandlerInfoList.get(i).getSourceAddress());

			Boolean sameTarget = newList.get(i).getClickTargetRate()
					.equals(sessionHandlerInfoList.get(i).getClickTargetRate());

			Boolean approxSamePing = Math.abs(newList.get(i).getPing()
					- sessionHandlerInfoList.get(i).getPing()) < 0.5d;

			// If they aren't equal, we need to return false
			Boolean theyAreEqual = sameClient && sameSource && sameTarget
					&& approxSamePing;
			if (!theyAreEqual) {
				return false;
			}
		}

		// Since we have not returned yet, they are all equal(approx)
		return true;

	}

	/**
	 * This method will based on the sessions in the markedSessions list
	 * generate a list of {@link SessionHandlerInfo} objects.<br>
	 * These objects will represent the session information relevant to the
	 * handler <br>
	 * 
	 * @return list of information representing the wanted sessions
	 */
	private List<SessionHandlerInfo> getNewListOfSessionHandlerInfo() {

		// The list
		List<SessionHandlerInfo> sessionHandlerInfoList = new ArrayList<SessionHandlerInfo>();

		for (ServiceSession session : markedSessions) {

			// Retrieve the ping for this session
			Double sessionPing = session.getPing();
			if (sessionPing == -1d || session.getClickTargetVideoBitRate() == 0) {
				/*
				 * If ping is not obtained or the user has not set a target rate
				 * this session should not be written to the Click handler
				 */
				continue;
			}

			List<String> sourceHostAddresses = session.getSourceHostAddresses();
			// Iterate over all possible source IP's. Create entry for each
			for (int i = 0; i < sourceHostAddresses.size(); i++) {

				// Get parameters
				String clientIp = session.getClientAddress().getHostAddress();
				String sourceIp = sourceHostAddresses.get(i);
				Double ping = sessionPing;
				Integer clickTargetRate = session.getClickTargetVideoBitRate();

				// Create the object
				SessionHandlerInfo handlerInfo = new SessionHandlerInfo(
						clientIp, sourceIp, ping, clickTargetRate);

				/*
				 * We cannot have two sessions with equal source/client ip. If
				 * this occurs, don't add the object. It will cause the two
				 * streams to suffer under a common bandwidth limit. But we
				 * don't want to crash Click
				 */
				Boolean duplicate = false;
				for (SessionHandlerInfo shi : sessionHandlerInfoList) {
					Boolean sameClient = shi.getClientAddress().equals(
							handlerInfo.getClientAddress());
					Boolean sameSource = shi.getSourceAddress().equals(
							handlerInfo.getSourceAddress());

					if (sameClient && sameSource) {
						System.out.println("Found duplicate");
						duplicate = true;
					}
				}

				if (!duplicate) {
					sessionHandlerInfoList.add(handlerInfo);
				}
			}
		}

		// And return
		return sessionHandlerInfoList;
	}

	/**
	 * A neat little method to clean and refactor the code. Will be called if
	 * the listHasChanged flag i set.Should then update the list we are working
	 * on and set the flag false. We set the updatedList to null, so that if
	 * anyone ever uses that list they will know it's wrong
	 */
	private void doListHasChangedRoutine() {
		markedSessions = updatedSessions;
		updatedSessions = null;
		listHasChanged = false;
	}

	/**
	 * Checks if the session has changed targetVideoBitrate since last time we
	 * read it.
	 * 
	 * @param markedSession
	 *            to retrieve oldTarget and Target from
	 * @return true if changed, false otherwise
	 */
	private boolean isTargetVideoBitrateChanged(ServiceSession markedSession) {
		return markedSession.getTargetVideoBitrate() != markedSession
				.getOldTargetVideoBitRate();
	}

	/**
	 * Takes a path and text as paramters. Writes the text to the given path.
	 * 
	 * @param filename
	 *            path to file
	 * @param text
	 *            String to write
	 */
	private void writeFile(String filename, String text) {

		try {

			FileWriter writer = new FileWriter(filename);

			writer.write(text);
			writer.close();

		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
