package no.ntnu.item.router;

import java.net.InetAddress;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;

import no.ntnu.item.router.utilities.SessionPinger;

/**
 * Domain object that represents a service session known to the router.<br>
 * It contains and maintains a {@link List} of {@link SilverlightMeasurement}
 * objects belonging to this {@link ServiceSession} instance. A service session
 * is in all simplicity the information collected from a video stream.
 * 
 * @author Kristian Haugene
 * 
 */
public class ServiceSession {

	/**
	 * The {@link List} of {@link SilverlightMeasurement} objects from this
	 * service
	 */
	private List<SilverlightMeasurement> measurements;

	/**
	 * Holds an integer set by the user, indicating if the user wants to affect
	 * it
	 */
	private Integer targetVideoBitRate;

	/**
	 * Holds the old targetVideoBitRate, used by {@link ActionReasoner} to
	 * decide if it needs to reload click rate calculations
	 */
	private Integer oldTargetVideoBitRate;

	/** For timing purposes, we want to know when the target rate was set */
	private Long targetRateSetTime;

	/**
	 * The current target rate set to click, we adjust this independent of user
	 * pref
	 */
	private Integer clickTargetVideoBitRate;

	/**
	 * This is a separate thread that will obtain ping measurements for the
	 * stream
	 */
	private SessionPinger sessionPinger;

	/** Holds the addresses our host can be */
	private List<String> sourceHostAddresses;

	/**
	 * Constructor
	 */
	public ServiceSession() {
		measurements = new ArrayList<SilverlightMeasurement>();
		sourceHostAddresses = new ArrayList<String>();
		setTargetVideoBitrate(0);
		setClickTargetVideoBitRate(0);
		setOldTargetVideoBitRate(0);
		sessionPinger = new SessionPinger(this);

	}

	/** Returns the current list of {@link SilverlightMeasurement} */
	public List<SilverlightMeasurement> getMeasurements() {
		return measurements;
	}

	/** Returns the latest measurement registered to this {@link ServiceSession} */
	public SilverlightMeasurement getLatestMeasurement() {
		return measurements.get(measurements.size() - 1);
	}

	/** Method for registering a {@link SilverlightMeasurement} to a service */
	public void addMeasurementToService(SilverlightMeasurement sm) {

		measurements.add(sm);
		System.out.println("New measurement from " + sm.getClientId());
	}

	/**
	 * This method calls the ping thread associated with this object and
	 * retrieves a ping
	 * 
	 * @return ping in ms with 2 decimals from router to streaming server
	 */
	public Double getPing() {

		// Get ping
		Double ping = sessionPinger.getPing();

		// Round to 2 decimals and return
		DecimalFormat twoDForm = new DecimalFormat("#.##");
		return Double.valueOf(twoDForm.format(ping));

	}

	/**
	 * Returns the silverlight client runtime from the latest measurement
	 * 
	 * @return silverlight client runtime
	 */
	public Integer getLatestTimeIncrement() {
		return getLatestMeasurement().getTimeIncrement().intValue();
	}

	/**
	 * Public method to retrieve the Silverlight media source IP addresses. The
	 * server is set as a dns name that might resolve differently from time to
	 * time. Return all values it has resolved to.
	 * 
	 * @return List of String IP entries host has resolved to
	 */
	public List<String> getSourceHostAddresses() {

		// Get the latest source address
		String latestSourceAddr = getLatestMeasurement().getSourceIpAndPort()
				.getAddress().getHostAddress();

		// If it does not exist in the list, update the list
		if (!sourceHostAddresses.contains(latestSourceAddr)) {
			sourceHostAddresses.add(latestSourceAddr);
		}

		// return the list
		return sourceHostAddresses;
	}

	/**
	 * Public method to retrieve the Silverlight client IP address
	 * 
	 * @return the client IP retrieved from the latest
	 *         {@link SilverlightMeasurement}
	 */
	public InetAddress getClientAddress() {
		return getLatestMeasurement().getMyIp();
	}

	/**
	 * A method that returns an identifier for this {@link ServiceSession}. A
	 * session is a set of measurements from the same client. The clientId is
	 * therefore a unique identifier of a {@link ServiceSession}.
	 */
	public String getId() {
		return measurements.get(0).getClientId();
	}

	public Integer getTargetVideoBitrate() {
		return targetVideoBitRate;
	}

	public Integer getOldTargetVideoBitRate() {
		return oldTargetVideoBitRate;
	}

	/**
	 * Sets the targetVideoBitrate for this session. Also updates the
	 * targetRateSetTime field to the current system time.
	 * 
	 * @param targetVideoBitrate
	 *            the wanted target rate
	 */
	public void setTargetVideoBitrate(Integer targetVideoBitrate) {
		this.targetVideoBitRate = targetVideoBitrate;

		// Update the time the rate was set. If rate is set to 0, so is time.
		if (targetVideoBitrate == 0) {
			this.targetRateSetTime = 0L;
		} else {
			this.targetRateSetTime = System.currentTimeMillis();
		}
	}

	/**
	 * Sets the oldTargetVideoBitrate for this session. Is used by
	 * {@link ActionReasoner} when it has read a new value and writes it here
	 * for further reference
	 * 
	 * @param oldTargetVideoBitRate to store
	 */
	public void setOldTargetVideoBitRate(Integer oldTargetVideoBitRate) {
		this.oldTargetVideoBitRate = oldTargetVideoBitRate;
	}

	public Long getTargetRateSetTime() {
		return targetRateSetTime;
	}

	public void setClickTargetVideoBitRate(Integer clickTargetVideoBitRate) {
		this.clickTargetVideoBitRate = clickTargetVideoBitRate;
	}

	public Integer getClickTargetVideoBitRate() {
		return clickTargetVideoBitRate;
	}

}
