package no.ntnu.item.router;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import no.ntnu.item.router.utilities.ActionPlaneSessions;
import no.ntnu.item.router.utilities.GuiXmlWriter;
import no.ntnu.item.router.utilities.Parser;
import no.ntnu.item.router.utilities.ServiceSessionBankCleaner;

/**
 * This is a registry for all service sessions known to the router.<br>
 * It contains functionality for handling the {@link ServiceSession} objects and
 * sorting incoming {@link SilverlightMeasurement} objects.
 * 
 * @author Kristian Haugene
 * 
 */
public class ServiceSessionBank {

	/** List of ongoing service sessions known to the Router */
	private List<ServiceSession> serviceSessions;

	/** Holds the {@link Parser} */
	private Parser parser;

	/** The {@link GuiXmlWriter} that is to be called when changes occur */
	private GuiXmlWriter guiXmlWriter;

	/** The {@link ActionPlaneSessions} that will keep track of AP sessions */
	private ActionPlaneSessions actionPlaneSessions;

	/**
	 * Constructor
	 * 
	 * @param parser
	 */
	public ServiceSessionBank(Parser parser, GuiXmlWriter guiXmlWriter,
			ActionPlaneSessions actionPlaneSessions) {
		this.parser = parser;
		serviceSessions = new ArrayList<ServiceSession>();
		this.guiXmlWriter = guiXmlWriter;
		this.actionPlaneSessions = actionPlaneSessions;

		// ServiceSessionBankCleaner watches for inactive serviceSessions
		new ServiceSessionBankCleaner(this);

		// Update, tell all that the list is empty
		updateGuiXmlAndApSessions();
	}

	/**
	 * Register a new {@link SilverlightMeasurement}
	 * 
	 * @param measurement
	 *            string representing a {@link SilverlightMeasurement}
	 * @return the clientId this measurement relates to
	 */
	public String registerMeasurement(String measurement) {

		// Parse the string to a SilverlightMeasurement, and add it
		SilverlightMeasurement sm = parser
				.getSilverlightMeasurementFromString(measurement);

		// null or empty measurement cause the parser to return null
		if (sm == null) {
			return null;
		}
		addMeasurement(sm);

		// We have changed list of sessions
		updateGuiXmlAndApSessions();
		
		return sm.getClientId();
		
	}

	/**
	 * Update user preferences
	 * 
	 * @param userPreferences
	 *            string representation
	 */
	public void registerUserPreferences(String userPreferences) {

		// Parse into a map of key,value pairs and convert to set
		Set<Entry<String, Integer>> userPreferencesSet = parser
				.getUserPreferencesMapFromString(userPreferences).entrySet();

		// Iterate the set
		Iterator<Entry<String, Integer>> iterator = userPreferencesSet
				.iterator();
		while (iterator.hasNext()) {
			Map.Entry<String, Integer> mapEntry = iterator.next();
			setTargetBitRateToSessionGivenById(mapEntry.getKey(),
					mapEntry.getValue());
		}

		// We have changed list of sessions
		updateGuiXmlAndApSessions();
	}

	/**
	 * Method that calls the {@link GuiXmlWriter} to write the current state.<br>
	 * Should be called when changes occur in the lists.<br>
	 * Also notifies the {@link ActionPlaneSessions} that a change has occured
	 */
	public void updateGuiXmlAndApSessions() {
		guiXmlWriter.writeXml(this);
		actionPlaneSessions.reportChangeInServiceSessionList(serviceSessions);
	}

	/**
	 * Returns the current list of {@link ServiceSession} objects.
	 * 
	 * @return current list of {@link ServiceSession} objects
	 */
	public List<ServiceSession> getServiceSessions() {
		return serviceSessions;
	}

	/**
	 * Method for adding a {@link SilverlightMeasurement} to the bank
	 * 
	 * @param sm
	 *            {@link SilverlightMeasurement} to add
	 */
	private void addMeasurement(SilverlightMeasurement sm) {

		// Should it be added to an existing service session?
		if (serviceSessionWithSameIdExistsInBank(sm)) {
			addMeasurementToExistingServiceSession(sm);
		} else {
			createNewServiceSessionAndAddMeasurement(sm);
		}
	}

	/**
	 * Given that the measurement is not from a new {@link ServiceSession}, adds
	 * the {@link SilverlightMeasurement} to the correct {@link ServiceSession}
	 * 
	 * @param sm
	 *            {@link SilverlightMeasurement} to add to existing
	 *            {@link ServiceSession}
	 */
	private void addMeasurementToExistingServiceSession(
			SilverlightMeasurement sm) {

		// Iterate through to the correct ServiceSession, add measurement
		for (ServiceSession session : serviceSessions) {
			if (measurementAndServiceSessionHaveSameId(session, sm)) {
				session.addMeasurementToService(sm);
			}
		}

	}

	/**
	 * Creates a new {@link ServiceSession} and adds the
	 * {@link SilverlightMeasurement} to it. Also adds the session to the
	 * {@link ServiceSessionBank} list of sessions.
	 * 
	 * @param sm
	 *            {@link SilverlightMeasurement} from a new session
	 */
	private void createNewServiceSessionAndAddMeasurement(
			SilverlightMeasurement sm) {
		System.out.println("New session with Id: " + sm.getClientId());

		// Create a new ServiceSession and add the measurement
		ServiceSession session = new ServiceSession();
		session.addMeasurementToService(sm);

		// Now add the ServiceSession to the ServiceBank serviceSessions list
		serviceSessions.add(session);
	}

	/**
	 * Simple method for checking if a {@link SilverlightMeasurement} has an Id
	 * that is equal to a already registered {@link ServiceSession}
	 * 
	 * @param measurement
	 *            {@link SilverlightMeasurement} to check for
	 * @return true if serviceSession list contains a session with this Id.
	 *         False otherwise.
	 */
	private boolean serviceSessionWithSameIdExistsInBank(
			SilverlightMeasurement measurement) {

		// If a session has corresponding Id, return true
		for (ServiceSession session : serviceSessions) {
			if (measurementAndServiceSessionHaveSameId(session, measurement)) {
				return true;
			}
		}

		// Did not find corresponding Id, return false
		return false;
	}

	/**
	 * Checks if the given {@link ServiceSession} and
	 * {@link SilverlightMeasurement} have the same Id
	 * 
	 * @param session
	 * @param measurement
	 * @return true if equal, false otherwise
	 */
	private boolean measurementAndServiceSessionHaveSameId(
			ServiceSession session, SilverlightMeasurement measurement) {

		// Check if Id corresponds
		if (session.getId().equals(measurement.getClientId())) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Takes cliendId and targetBitRate as parameter.<br>
	 * Will set the targetBitRate to the {@link ServiceSession} identified by
	 * clientId
	 * 
	 * @param clientId
	 *            the session id
	 * @param targetBitRate
	 *            wanted bit rate
	 */
	private void setTargetBitRateToSessionGivenById(String clientId,
			Integer targetBitRate) {

		for (ServiceSession session : serviceSessions) {
			if (session.getId().equals(clientId)) {
				session.setTargetVideoBitrate(targetBitRate);
			}
		}

	}

}
