package no.ntnu.item.router.utilities;

import java.util.ArrayList;
import java.util.List;

import no.ntnu.item.router.ActionReasoner;
import no.ntnu.item.router.ServiceSession;
import no.ntnu.item.router.ServiceSessionBank;

/**
 * The responsibility of this class is to make a list of {@link ServiceSession}
 * that the user has set a target bit rate for. These are the
 * {@link ServiceSession} objects that AP should care about. Every time
 * {@link ServiceSessionBank} is updated it needs to run through the elements.
 * It then selects the ones that the router should perform action on/for.<br>
 * This list should then be communicated to the Thread actually responsible<br>
 * for continuous communication with the AP layer.
 * 
 * @author Kristian Haugene
 * 
 */
public class ActionPlaneSessions {

	/** The list of {@link ServiceSession} objects relevant to AP */
	private List<ServiceSession> markedSessions = new ArrayList<ServiceSession>();

	/** The reasoner to call when the list of sessions is changed */
	private ActionReasoner actionReasoner;

	/**
	 * Constructor. Takes a {@link ActionReasoner} as parameter.<br>
	 * {@link ActionReasoner} will be called when changes markedSessions occur
	 * 
	 * @param actionReasoner
	 *            to call when changes occur
	 */
	public ActionPlaneSessions(ActionReasoner actionReasoner) {
		this.actionReasoner = actionReasoner;
	}

	/**
	 * This method should be called when a change occurs in
	 * {@link ServiceSessionBank}. When changes occur, the list obtained by
	 * might have to be updated accordingly.
	 * 
	 * @param updatedList
	 *            the new list of {@link ServiceSession} objects
	 */
	public void reportChangeInServiceSessionList(
			List<ServiceSession> updatedList) {

		// Retrieve the sessions marked with target bit rate
		List<ServiceSession> apSessions = getApSessionsFromNewList(updatedList);

		// Check if the list has changed
		if (areServiceSessionsChanged(markedSessions, apSessions)) {

			// It has, remember the new list
			markedSessions = apSessions;

			// Update actionReasoner
			actionReasoner.updateListOfMarkedSessions(markedSessions);

		} else {
			/*
			 * No change. Therefore no need to do anything. The old list
			 * contains object reference to the current set of ServiceSessions
			 * and will get information from them
			 */
		}
	}

	/**
	 * Extracts a list of {@link ServiceSession} objects that have a target bit
	 * rate set
	 * 
	 * @param updatedList
	 *            the list to iterate
	 * @return the subset that have target bit rate set
	 */
	private List<ServiceSession> getApSessionsFromNewList(
			List<ServiceSession> updatedList) {

		List<ServiceSession> sessionsWithTargetBitRate = new ArrayList<ServiceSession>();

		for (ServiceSession session : updatedList) {
			if (session.getTargetVideoBitrate() != 0) {
				sessionsWithTargetBitRate.add(session);
			}
		}

		return sessionsWithTargetBitRate;

	}

	/**
	 * This method checks two lists of {@link ServiceSession} objects to see if
	 * they contain the same elements
	 * 
	 * @param oldList
	 *            the first list
	 * @param newList
	 *            the new list
	 * @return true if the two lists have an equal number of objects and the
	 *         objects are the same in both lists. False otherwise
	 */
	private Boolean areServiceSessionsChanged(List<ServiceSession> oldList,
			List<ServiceSession> newList) {

		// The lists must be of equal size to be equal
		if (oldList.size() != newList.size()) {
			return true;
		}

		/*
		 * For each oldSession, find corresponding newSession. If found,
		 * increment foundMatches.
		 */
		int foundMatches = 0;
		for (ServiceSession oldSession : oldList) {
			for (ServiceSession newSession : newList) {
				if (oldSession.getId().equals(newSession.getId())) {
					foundMatches++;
				}
			}
		}

		/*
		 * If the number of matches found is not equal to list size there have
		 * been change
		 */
		if (foundMatches != oldList.size()) {
			return true;
		}

		// Lists are of equal size and contain the same elements. No change
		return false;
	}

}