package no.ntnu.item.router.utilities;

import static no.ntnu.item.router.Constants.WEB_GUI_XML_PATH;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import no.ntnu.item.router.ServiceSession;
import no.ntnu.item.router.ServiceSessionBank;
import no.ntnu.item.router.SilverlightMeasurement;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sun.org.apache.xml.internal.serialize.XMLSerializer;

/**
 * This is a class with functionality for informing the user about the current
 * network situation. When this class is called, it will write certain
 * parameters to file. This file will be read by the apache server that hosts
 * the Web UI.
 * 
 * @author Kristian Haugene
 * 
 */
public class GuiXmlWriter {

	public void writeXml(ServiceSessionBank serviceSessionBank) {

		try {
			// first of all we request our DOM-implementation:
			DocumentBuilderFactory factory = DocumentBuilderFactory
					.newInstance();
			// then we have to create document-loader:
			DocumentBuilder loader = factory.newDocumentBuilder();

			// Creating a new DOM-document...
			Document document = loader.newDocument();

			// create root-element
			Element root = document.createElement("root");
			document.appendChild(root);

			// For each ServiceSession, add data to xml root
			List<ServiceSession> serviceSessions = serviceSessionBank
					.getServiceSessions();
			for (ServiceSession session : serviceSessions) {
				addServiceSession(session, document, root);
			}

			// use specific Xerces class to write DOM-data to a file:
			XMLSerializer serializer = new XMLSerializer();
			serializer.setOutputCharStream(new java.io.FileWriter(
					WEB_GUI_XML_PATH));
			serializer.serialize(document);

		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}

	/**
	 * Method for adding a serviceSession to the root element
	 * 
	 * @param serviceSession
	 *            the session to add
	 * @param document
	 *            for element creation
	 * @param root
	 *            to append the element to
	 */
	private void addServiceSession(ServiceSession serviceSession,
			Document document, Element root) {

		SilverlightMeasurement latestMeasurement = serviceSession
				.getLatestMeasurement();

		// Create the element
		Element sessionElem = document.createElement("serviceSession");

		// Set attributes
		addChildElement(document, sessionElem, "clientId",
				serviceSession.getId());
		addChildElement(document, sessionElem, "videoDownloadBitrate",
				latestMeasurement.getVideoDownloadBitrate().toString());
		addChildElement(document, sessionElem, "routerTimeStamp",
				latestMeasurement.getRouterTimeStamp().toString());
		addChildElement(document, sessionElem, "playState", latestMeasurement
				.getPlayState().toString());
		addChildElement(document, sessionElem, "fullScreen", latestMeasurement
				.getFullScreen().toString());
		addChildElement(document, sessionElem, "avgProcessorLoad",
				latestMeasurement.getAvgProcessorLoad().toString());
		addChildElement(document, sessionElem, "event",
				latestMeasurement.getEvent());
		addChildElement(document, sessionElem, "myIp", latestMeasurement
				.getMyIp().getHostAddress());
		addChildElement(document, sessionElem, "targetVideoRate",
				serviceSession.getTargetVideoBitrate().toString());
		addChildElement(document, sessionElem, "rtt", serviceSession.getPing()
				.toString());
		addChildElement(document, sessionElem, "sessionTime", serviceSession
				.getLatestTimeIncrement().toString());

		// Currently not in use, but could be useful?
		// String humanReadableTimeStamp =
		// getHumanReadableTimeStamp(latestMeasurement);

		addAvaliableVideoBitrates(latestMeasurement, document, sessionElem);
		addBitRateStatistics(serviceSession, document, sessionElem);

		// Add it to the root
		root.appendChild(sessionElem);
	}

	/**
	 * A method for adding all the avaliable video bit rates to the
	 * sessionElement.
	 * 
	 * @param latestMeasurement
	 *            to retrieve list of bit rates
	 * @param document
	 *            for element creation
	 * @param sessionElem
	 *            where the bit rates should be added
	 */
	private void addAvaliableVideoBitrates(
			SilverlightMeasurement latestMeasurement, Document document,
			Element sessionElem) {

		List<Integer> bitrateList = latestMeasurement
				.getAvaliableVideoBitrates();

		// Create the avaliableVideoBitrates node
		Element avaliableVideoBitrates = document
				.createElement("avaliableVideoBitrates");

		// Append all the bitrates
		for (Integer bitRate : bitrateList) {
			addChildElement(document, avaliableVideoBitrates, "bitRate",
					bitRate.toString());
		}

		// Finally append the list to the sessionElement
		sessionElem.appendChild(avaliableVideoBitrates);

	}

	/**
	 * Method for adding bit rate statistics to the sessionElem
	 * 
	 * @param serviceSession
	 *            to retrieve the measurements from
	 * @param document
	 *            for element creation
	 * @param sessionElem
	 *            where the statistics are to be added
	 */
	private void addBitRateStatistics(ServiceSession serviceSession,
			Document document, Element sessionElem) {

		// Create the bitRateStatistics node
		Element bitRateStatistics = document.createElement("bitRateStatistics");

		// Retrieve list of measurements
		List<SilverlightMeasurement> measurements = serviceSession
				.getMeasurements();

		// We need to iterate through the session measurements
		for (SilverlightMeasurement measurement : measurements) {

			// Create an entry and add values from the measurement
			Element entry = document.createElement("entry");
			addChildElement(document, entry, "bitRate", measurement
					.getVideoDownloadBitrate().toString());
			addChildElement(document, entry, "timeStamp", measurement
					.getRouterTimeStamp().toString());

			// Finally append this entry to the bitrateStatistics
			bitRateStatistics.appendChild(entry);
		}

		// And finally append the statistics to the sessionElem
		sessionElem.appendChild(bitRateStatistics);

	}

	/**
	 * Method that makes and adds a child element to a already existing parent.
	 * 
	 * @param document
	 *            for creation of elements
	 * @param parent
	 *            parent node
	 * @param name
	 *            the name of the new child
	 * @param value
	 *            value of the child elements
	 */
	private void addChildElement(Document document, Element parent,
			String name, String value) {

		Element elem = document.createElement(name);
		elem.setTextContent(value);

		parent.appendChild(elem);

	}

	/**
	 * This method extracts a human readable time stamp on the form hh:mm:ss
	 * from a {@link SilverlightMeasurement}
	 * 
	 * TODO Remove this method if it will never be used, also remove commented
	 * line where this method is used earlier in the code
	 * 
	 * @param latestMeasurement
	 *            to read the time stamp from
	 * @return a string representation on the form of hh:mm:ss
	 */
	@SuppressWarnings("unused")
	private String getHumanReadableTimeStamp(
			SilverlightMeasurement latestMeasurement) {

		String humanReadableTime = "";
		Calendar cal = new GregorianCalendar();

		// Set time to the current timeStamp and read hour, minute and second
		cal.setTime(new Date(latestMeasurement.getRouterTimeStamp()));
		Integer hour = cal.get(Calendar.HOUR_OF_DAY);
		Integer minute = cal.get(Calendar.MINUTE);
		Integer second = cal.get(Calendar.SECOND);

		// make hh:mm:ss syntax, add hours
		humanReadableTime += hour + ":";

		// Add minutes, make sure that 5 is 05
		if (minute < 10) {
			humanReadableTime += "0" + minute + ":";
		} else {
			humanReadableTime += minute + ":";
		}

		// Add seconds, make sure that 5 is 05
		if (second < 10) {
			humanReadableTime += "0" + second;
		} else {
			humanReadableTime += second;
		}

		return humanReadableTime;
	}

}
