package no.ntnu.item.router.utilities;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import no.ntnu.item.router.SilverlightMeasurement;
import no.ntnu.item.router.enums.PlayState;

/**
 * A class containing parsers. To externalize the parsing of objects
 * 
 * @author Kristian Haugene
 * 
 */
public class Parser {
	

	public SilverlightMeasurement getSilverlightMeasurementFromString(
			String measurement) {
		
		// Init return value
		SilverlightMeasurement sm = null;
		
		// If no input, return null
		if(measurement == null || measurement.equals("")){
			return null;
		}
		
		// Split the measurement at POST message "&" character
		String[] parameters = measurement.split("&");
		
		// Parameters are given as event=InitialInformation, remove text until =
		for(int i = 0 ; i < parameters.length ; i++){
			parameters[i] = removeTextBeforeAndIncludingEqualityCharacter(parameters[i]);
		}
		
		try {
		
			// Extract the parameters, give them names
			String event = parameters[0].trim();
			String eventMessage = parameters[1].trim();
			PlayState playState = getPlayStateOrDefaultToUnknown(parameters[2]);
			Boolean fullScreen = Boolean.valueOf(parameters[3].trim());
			String clientId = parameters[4].trim();
			Double avgProcessorLoad = Double.valueOf(parameters[5].trim().replace(',', '.'));
			Double avgProcessLoad = Double.valueOf(parameters[6].trim().replace(',', '.'));
			Integer videoDownloadBitrate = getDownloadBitRateOrZero(parameters[7]);
			List<Integer> avaliableVideoBitrates = buildListFromString(parameters[8].trim());
			InetAddress myIp = InetAddress.getByName(parameters[9].trim());
			String ispAs = parameters[10].trim();
			InetSocketAddress sourceIpAndPort = getAddressFromString(parameters[11].trim());
			Double timeIncrement = Double.valueOf(parameters[12].trim().replace(',', '.'));
			
			// And collect a timestamp
			Long timeStamp = System.currentTimeMillis();

			sm = new SilverlightMeasurement(event, eventMessage, clientId,
					ispAs, playState, fullScreen, avgProcessorLoad,
					avgProcessLoad, timeIncrement, videoDownloadBitrate,
					avaliableVideoBitrates, myIp, sourceIpAndPort, timeStamp);

		} catch (Exception e) {
			System.out.println("Caught exception in parsing measurement");
			e.printStackTrace();
		}
		
		// We had an exception, return the null value
		return sm;
		
	}

	/**
	 * Helper method to extract PlayState. If no PlayState is recognizer by our
	 * enum, it will default to UNKNOWN
	 * 
	 * @param playStateString
	 *            the playState represented by a string
	 * @return corresponding {@link PlayState} object
	 */
	private PlayState getPlayStateOrDefaultToUnknown(String playStateString) {
		PlayState playState;
		try {
			playState = PlayState.valueOf(playStateString.trim().toUpperCase());
		} catch (Exception e) {
			playState = PlayState.UNKNOWN;
		}
		return playState;
	}

	/**
	 * Helper method to extract videoDownloadBitrate. If no rate is given it
	 * will default to 0
	 * 
	 * @param bitRateString
	 *            as string
	 * @return videoDownloadBitrate as Integer, or 0 if a parse-error occurs
	 */
	private Integer getDownloadBitRateOrZero(String bitRateString) {
		Integer videoDownloadBitrate;
		try {
			videoDownloadBitrate = Integer.valueOf(bitRateString.trim());
		} catch (Exception e) {
			videoDownloadBitrate = 0;
		}
		return videoDownloadBitrate;
	}

	/**
	 * Helper method. Avaliable bitrates is a {@link String} formatted like
	 * "bitRate - bitRate - bitRate". We want to take all the bitRates and put
	 * them in a {@link List}
	 * 
	 * @param bitrates String containing "-" separated bitRates
	 * @return the bitRates in a {@link List}
	 */
	private List<Integer> buildListFromString(String bitrates) {

		List<Integer> bitRateList = new ArrayList<Integer>();
		
		// Add a "not set" value
		bitRateList.add(0);
		
		// First split them
		String[] rateArray = bitrates.split("-");
		
		// Then add convert them to integer and add them to the list
		for(String bitRate : rateArray){
			bitRateList.add(Integer.valueOf(bitRate.trim()));
		}
		
		// And return
		return bitRateList;
	}

	/**
	 * Helper method. Takes a string parameter and constructs a
	 * {@link InetSocketAddress}
	 * 
	 * @param input
	 *            string on the form ip:port
	 * @return {@link InetSocketAddress} representing the ip:port given
	 */
	private InetSocketAddress getAddressFromString(String input) {

		// The input string has the form ip:port
		String[] adrArray = input.split(":");

		String address = adrArray[0].trim();
		Integer port = Integer.valueOf(adrArray[1].trim());

		// Make a socketAddress and return
		return new InetSocketAddress(address, port);

	}

	/**
	 * Helper method. Given that parameters are written like this:<br>
	 * "event=InitialInformation". This method will return what is after the "="
	 * sign.
	 * 
	 * @param parameter
	 *            as received in the POST message
	 * @return the value
	 */
	private String removeTextBeforeAndIncludingEqualityCharacter(
			String parameter) {
		String toReturn = parameter.substring(parameter.indexOf("=") + 1);
		return toReturn;
	}

	/**
	 * Takes user preferences on the form "clientId=bitRate&clientId2=bitRate2"<br>
	 * I.e, regular HTTP POST way of relaying information. Will return them as a<br>
	 * {@link HashMap} of key,value pairs like clientId,bitRate
	 * 
	 * @param userPreferences
	 *            string representation of cliendID and bitRate pairs
	 * @return the clientId,bitRate pairs in a {@link HashMap}
	 */
	public HashMap<String, Integer> getUserPreferencesMapFromString(String userPreferences) {
		
		HashMap<String, Integer> hashMap = new HashMap<String, Integer>();
		
		// If no input, return empty Map
		if(userPreferences == null || userPreferences.equals("")){
			return hashMap;
		}
		
		// Split the preferences
		String[] preferenceArray = userPreferences.split("&");
		
		// For each clientId=xxx paid, split at "=" and add as Map<K,V> pair
		for(int i = 0; i< preferenceArray.length; i++){
			String[] aPreference = preferenceArray[i].split("=");

			// If the target value or clientId is "", jump to next preference
			if(aPreference[0].isEmpty() || aPreference[1].isEmpty()){
				continue;
			}
			
			String cliendId = aPreference[0];			
			Integer targetBitRate = Integer.parseInt(aPreference[1]);
			hashMap.put(cliendId, targetBitRate);
		}
		
		return hashMap;
		
	}

}
