package no.ntnu.item.router.utilities;

import java.io.BufferedReader;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetSocketAddress;

import no.ntnu.item.router.ServiceSession;

/**
 * This class will obtain ping measurements for the given {@link ServiceSession}
 * 
 * @author Kristian Haugene
 * 
 */
public class SessionPinger extends Thread {

	/** The {@link ServiceSession} for which this class will sample ping */
	private ServiceSession serviceSession;

	/** Holds the 5 last ping values for computing average */
	Double[] pingValues = { 0d, 0d, 0d, 0d, 0d };

	/**
	 * Constructor. Set local variable, start thread
	 * 
	 * @param session
	 *            the session to ping for
	 */
	public SessionPinger(ServiceSession session) {

		// Set the session for which to ping
		serviceSession = session;

		this.start();
	}

	/**
	 * The running method
	 */
	public void run() {

		while (true) {
			try {
				Thread.sleep(2000);
			} catch (Exception e) {
			}
			
			/*
			 * TODO Here's a small hack
			 * Will force the application to resolve the hostname each second
			 * Then we will get a good list of possible source addresses sooner
			 */
			serviceSession.getSourceHostAddresses();

			InetSocketAddress addr = serviceSession.getLatestMeasurement()
					.getSourceIpAndPort();
			addPingValue(pingAddress(addr));
		}

	}

	/**
	 * Uses the array of ping values.<br>
	 * Summarizes and averages them.
	 * 
	 * @return ping as double
	 */
	public Double getPing() {

		if (haveNoPingValues()) {
			return -1d;
		}

		Double sum = 0d;

		// Sum them
		int summed = 0;
		for (int i = 0; i < pingValues.length; i++) {
			// If value is 0, don't add(init of array)
			if (pingValues[i] != 0d) {
				sum += pingValues[i];
				summed++;
			}
		}

		Double avgPing = sum / summed;

		return avgPing;

	}

	/**
	 * Updates the array of ping values with a new one
	 * 
	 * @param pingAddress
	 *            to add
	 */
	private void addPingValue(Double pingAddress) {
		
		// If input is 0, don't add
		if(pingAddress == 0){
			return;
		}

		// Move the last measurements 1 index down
		for (int i = 1; i < pingValues.length; i++) {
			pingValues[i - 1] = pingValues[i];
		}

		// and add value at the end of the array
		pingValues[pingValues.length - 1] = pingAddress;

	}

	/**
	 * /** Running ping command in linux. Will return the ping value
	 * 
	 * @param addr
	 *            address of the server to ping
	 * @return decimal ping value
	 */
	private Double pingAddress(InetSocketAddress addr) {
		String s = null;

		String result = "";
		Process p = null;
		try {

			String command = "ping -c 1 " + addr.getAddress().getHostAddress();

			// using the Runtime exec method:
			p = Runtime.getRuntime().exec(command);

			BufferedReader stdInput = new BufferedReader(new InputStreamReader(
					p.getInputStream()));

			BufferedReader stdError = new BufferedReader(new InputStreamReader(
					p.getErrorStream()));

			// read the output from the command
			// System.out.println("Here is the standard output of the command:\n");
			while ((s = stdInput.readLine()) != null) {
				// System.out.println(s);
				result += s;
			}

			// read any errors from the attempted command
			// System.out.println("Here is the standard error of the command (if any):\n");
			while ((s = stdError.readLine()) != null) {
				// System.out.println(s);
				result = "0";
			}

		} catch (IOException e) {
			System.out.println("exception happened - here's what I know: ");
			e.printStackTrace();
			System.exit(-1);
		} finally {

			// When all is done, close
			if (p != null) {
				close(p.getOutputStream());
				close(p.getInputStream());
				close(p.getErrorStream());
				p.destroy();
			}
		}

		// If an error occured, return 0
		if(result.equals("0")){
			return 0d;
		}
		
		// We got result. Substring to result, parse to Double, round and return
		String fromMsAndOut = result.substring(result.indexOf("time=") + 5);
		String alsoCutBehindMs = fromMsAndOut.substring(0,
				fromMsAndOut.indexOf("ms") - 1);
		
		Double pingWithDecimals;
		try {
			 pingWithDecimals = Double.valueOf(alsoCutBehindMs);
		} catch (Exception e) {
			pingWithDecimals = 0d;
		}


		return pingWithDecimals;

	}

	/**
	 * Checks if the pingValues array is empty(all 0)
	 * 
	 * @return true if all elements are 0, false otherwise
	 */
	private boolean haveNoPingValues() {

		for (int i = 0; i < pingValues.length; i++) {
			if (pingValues[i] != 0) {

				// We found a value different from 0, return false
				return false;
			}
		}

		// No value was != 0, return true
		return true;
	}

	/**
	 * Closes the object given as parameter
	 * 
	 * @param c
	 *            a closable object
	 */
	private static void close(Closeable c) {
		if (c != null) {
			try {
				c.close();
			} catch (IOException e) {
				// ignored
			}
		}
	}

}
