package no.ntnu.item.router.utilities;

import static java.net.HttpURLConnection.HTTP_OK;
import static no.ntnu.item.router.Constants.WEB_SERVER_PORT;
import static no.ntnu.item.router.Constants.WEB_SERVER_ADDR;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.URLDecoder;
import java.util.List;

import no.ntnu.item.router.ServiceSession;
import no.ntnu.item.router.ServiceSessionBank;

import com.sun.net.httpserver.Headers;
import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpHandler;
import com.sun.net.httpserver.HttpServer;

/**
 * This is the main class for a web server. We use sun web server. NB: You need
 * to change build path to include JRE library from java6-sun instead of the
 * openJdk
 * 
 * The class starts a web server and has a corresponding handler for every
 * context. The server will be run as a background thread.
 * 
 * @author Kristian Haugene
 */
public class WebServer {

	final InetSocketAddress addr;
	final HttpServer server;

	public WebServer(ServiceSessionBank serviceBank) throws IOException {

		// Create the InetSocketAddress to be used for connections
		addr = new InetSocketAddress(WEB_SERVER_ADDR, WEB_SERVER_PORT);

		// Create server on given address with 10 queue slots
		server = HttpServer.create(addr, 10);

		// Create a handler to respond with a clientaccesspolicy file that is
		// required by Silverlight
		server.createContext("/clientaccesspolicy",
				new ClientAccessPolicyHandler());

		// Create a handler to process POST messages of log info
		server.createContext("/receiveParams", new PostInfoHandler(serviceBank));

		// Create a handler to receive user preferences
		server.createContext("/userPreferences", new UserInputHandler(
				serviceBank));

		// Create a handler to process all else, just echo
		server.createContext("/", new EchoHandler());

		// And GO!
		server.start();
	}
}

/**
 * This is a {@link HttpHandler} for the reception of HTTP POST messages.<br>
 * The POST messages are sent from the Silverlight clients behind the router<br>
 * The messages are read and passed on to the {@link ServiceSessionBank}<br>
 * After processing the incoming request, a response is constructed and sent.
 * 
 * @author Kristian Haugene
 * 
 */
class PostInfoHandler implements HttpHandler {

	/** The {@link ServiceSessionBank} object to report measurements to */
	private ServiceSessionBank serviceBank;

	/**
	 * Constructor
	 * 
	 * @param serviceBank
	 */
	public PostInfoHandler(ServiceSessionBank serviceBank) {
		this.serviceBank = serviceBank;
	}

	public void handle(HttpExchange t) throws IOException {
		final InputStream is;
		final OutputStream os;
		StringBuilder buf;
		int b;
		final String request, response;

		buf = new StringBuilder();

		/*
		 * Get the request body and decode it. Regardless of what you are
		 * actually doing, it is apparently considered correct form to consume
		 * all the bytes from the InputStream. If you don't, closing the
		 * OutputStream will cause that to occur
		 */

		is = t.getRequestBody();

		while ((b = is.read()) != -1) {
			buf.append((char) b);
		}

		is.close();

		if (buf.length() > 0) {
			request = URLDecoder.decode(buf.toString(), "UTF-8");
		} else {
			request = null;
		}

		/*
		 * Register the measurement and construct our response:
		 */

//		System.out.println(request);
		String clientId = serviceBank.registerMeasurement(request);

//		String clientId = null;
		// Get the user target rate if set
		Integer userTargetRate = null;
		if (clientId != null) {
			userTargetRate = getUserTargetRate(clientId);
		} else {
			userTargetRate = 0;
		}

		buf = new StringBuilder();

		buf.append("Click22 server: PostInfo");
		buf.append(t.getRequestMethod() + " " + t.getRequestURI() + " "
				+ t.getProtocol() + "\n");

		buf.append("Received message at time: " + System.currentTimeMillis());

		// If user target rate is set. Give this information to silverlight
		buf.append("\n");

		buf.append("userRate=");
		buf.append(userTargetRate);

		response = buf.toString();

		/*
		 * And now send the response. We could have instead done this
		 * dynamically, using 0 as the response size (forcing chunked encoding)
		 * and writing the bytes of the response directly to the OutputStream,
		 * but building the String first allows us to know the exact length so
		 * we can send a response with a known size. Better :)
		 */

		t.sendResponseHeaders(HTTP_OK, response.length());

		os = t.getResponseBody();

		os.write(response.getBytes());

		/*
		 * And we're done!
		 */

		os.close();
		t.close();
	}

	/**
	 * This method will retrieve the user target rate if set
	 * 
	 * @param clientId
	 *            identifying the session to retireve rate from
	 * @return the target rate as integer, null if rate is not set
	 */
	private Integer getUserTargetRate(String clientId) {

		List<ServiceSession> sessionList = serviceBank.getServiceSessions();
		for (ServiceSession session : sessionList) {
			if (session.getId().equals(clientId)) {
				return session.getTargetVideoBitrate();
			}
		}

		return null;
	}
}

/**
 * This is a handler for the clientaccesspolicy context.<br>
 * It's only task is to return the clientaccesspolicy.xml when the server is
 * asked for it. This is a Silverlight thing, for security reasons our server
 * must allow the Silverlight application to read from it. It will ask for the
 * xml file.
 * 
 * @author Kristian Haugene
 */
class ClientAccessPolicyHandler implements HttpHandler {

	public void handle(HttpExchange t) throws IOException {
		final InputStream is;
		final OutputStream os;
		StringBuilder buf;
		int b;

		@SuppressWarnings("unused")
		final String request, response;

		buf = new StringBuilder();

		/*
		 * Get the request body and decode it. Regardless of what you are
		 * actually doing, it is apparently considered correct form to consume
		 * all the bytes from the InputStream. If you don't, closing the
		 * OutputStream will cause that to occur
		 */

		is = t.getRequestBody();

		while ((b = is.read()) != -1) {
			buf.append((char) b);
		}

		is.close();

		if (buf.length() > 0) {
			request = URLDecoder.decode(buf.toString(), "UTF-8");
		} else {
			request = null;
		}

		/*
		 * Construct our response:
		 */

		response = "<?xml version=\"1.0\" encoding=\"utf-8\"?><access-policy><cross-domain-access><policy><allow-from><domain uri=\"*\"/></allow-from><grant-to><resource path=\"/\" include-subpaths=\"true\" /></grant-to></policy></cross-domain-access></access-policy>";

		/*
		 * And now send the response. We could have instead done this
		 * dynamically, using 0 as the response size (forcing chunked encoding)
		 * and writing the bytes of the response directly to the OutputStream,
		 * but building the String first allows us to know the exact length so
		 * we can send a response with a known size. Better :)
		 */

		t.sendResponseHeaders(HTTP_OK, response.length());

		os = t.getResponseBody();

		os.write(response.getBytes());

		/*
		 * And we're done!
		 */

		os.close();
		t.close();
	}

}

/**
 * This is a handler for the user input.<br>
 * When a user makes changes in the Web UI, a message will be sent here.<br>
 * <br>
 * NOTE: Why did we do this over http? Why not make the UI write to the xml<br>
 * file where it reads the data? Because then the RouterControl would have to<br>
 * poll for changes in the file and compare to what it has locally.<br>
 * In addition, we might create a stand-alone control panel, which <br>
 * reads the xml and sends messages to the server with preferences<br>
 * 
 * @author Kristian Haugene
 * 
 */
class UserInputHandler implements HttpHandler {

	/** The {@link ServiceSessionBank} object to report measurements to */
	private ServiceSessionBank serviceBank;

	/**
	 * Constructor
	 * 
	 * @param serviceBank
	 */
	public UserInputHandler(ServiceSessionBank serviceBank) {
		this.serviceBank = serviceBank;
	}

	public void handle(HttpExchange t) throws IOException {
		final InputStream is;
		final OutputStream os;
		StringBuilder buf;
		int b;
		final String request, response;

		buf = new StringBuilder();

		/*
		 * Get the request body and decode it. Regardless of what you are
		 * actually doing, it is apparently considered correct form to consume
		 * all the bytes from the InputStream. If you don't, closing the
		 * OutputStream will cause that to occur
		 */

		is = t.getRequestBody();

		while ((b = is.read()) != -1) {
			buf.append((char) b);
		}

		is.close();

		if (buf.length() > 0) {
			request = URLDecoder.decode(buf.toString(), "UTF-8");
		} else {
			request = null;
		}

		/*
		 * The user preferences POST will contain 1 or more user preferences.
		 * After these there will follow a redirect location. We want to
		 * redirect to the given location and remove the info before parsing it
		 * as a user preference.
		 */
		String redirectLocation = getRedirectLocation(request);
		String requestStrippedOfRedirectParam = removeTrailingRedirectParameter(request);

		/*
		 * Register the measurement and construct our response:
		 */
		serviceBank.registerUserPreferences(requestStrippedOfRedirectParam);

		buf = new StringBuilder();

		// Html Content, with redirect to the given site
		buf.append("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">");
		buf.append("<html><head><title>Click22 UserPreferences Response</title>");
		buf.append("<meta http-equiv=\"REFRESH\" content=\"0;url=http://"
				+ redirectLocation + "\"></head>");
		buf.append("<body><p>Click22 server: UserInput</p>");
		buf.append("<p>Received message at time: " + System.currentTimeMillis());
		buf.append("</p></body></html>");

		response = buf.toString();

		/*
		 * And now send the response. We could have instead done this
		 * dynamically, using 0 as the response size (forcing chunked encoding)
		 * and writing the bytes of the response directly to the OutputStream,
		 * but building the String first allows us to know the exact length so
		 * we can send a response with a known size. Better :)
		 */

		t.sendResponseHeaders(HTTP_OK, response.length());

		os = t.getResponseBody();

		os.write(response.getBytes());

		/*
		 * And we're done!
		 */

		os.close();
		t.close();
	}

	/**
	 * Helper method. Will take a string and remove everything after and
	 * including "&redirect=" tag. This is done before parsing the input as user
	 * preferences
	 * 
	 * @param request
	 *            the http request body
	 * @return the request body where the redirect parameter is stripped
	 */
	private String removeTrailingRedirectParameter(String request) {

		// Now make a request string where the redirect=xxx is stripped
		return request.substring(0, request.indexOf("&redirect="));

	}

	/**
	 * Will examine the http request body and return the location given by the
	 * redirect parameter
	 * 
	 * @param request
	 *            to look for redirect address
	 * @return the address given as parameter
	 */
	private String getRedirectLocation(String request) {

		// Return redirect location
		return request.substring(request.indexOf("&redirect=") + 10,
				request.length());
	}
}

/**
 * This is a simple handler for testing purposes. Simply echoes back a hello
 * message<br>
 * And it echoes the http header that it received
 * 
 * @author Kristian Haugene
 * 
 */
class EchoHandler implements HttpHandler {

	public void handle(HttpExchange t) throws IOException {
		final InputStream is;
		final OutputStream os;
		StringBuilder buf;
		int b;
		final String request, response;

		buf = new StringBuilder();

		/*
		 * Get the request body and decode it. Regardless of what you are
		 * actually doing, it is apparently considered correct form to consume
		 * all the bytes from the InputStream. If you don't, closing the
		 * OutputStream will cause that to occur
		 */

		is = t.getRequestBody();

		while ((b = is.read()) != -1) {
			buf.append((char) b);
		}

		is.close();

		if (buf.length() > 0) {
			request = URLDecoder.decode(buf.toString(), "UTF-8");
		} else {
			request = null;
		}

		System.out.println("EchoServer: " + request);

		/*
		 * Construct our response:
		 */

		buf = new StringBuilder();
		buf.append("<html><head><title>Click22 server: EchoHandler</title></head><body>");
		buf.append("<p><pre>");
		buf.append(t.getRequestMethod() + " " + t.getRequestURI() + " "
				+ t.getProtocol() + "\n");

		/*
		 * Process the request headers. This is a bit involved due to the
		 * complexity arising from the fact that headers can be repeated.
		 */

		Headers headers = t.getRequestHeaders();

		for (String name : headers.keySet()) {
			List<String> values = headers.get(name);

			for (String value : values) {
				buf.append(name + ": " + value + "\n");
			}
		}

		/*
		 * If there was an actual body to the request, add it:
		 */

		if (request != null) {
			buf.append("\n");
			buf.append(request);
		}

		buf.append("</pre></p>");
		buf.append("</body></html>\n");

		response = buf.toString();

		/*
		 * And now send the response. We could have instead done this
		 * dynamically, using 0 as the response size (forcing chunked encoding)
		 * and writing the bytes of the response directly to the OutputStream,
		 * but building the String first allows us to know the exact length so
		 * we can send a response with a known size. Better :)
		 */

		t.sendResponseHeaders(HTTP_OK, response.length());

		os = t.getResponseBody();

		os.write(response.getBytes());

		// System.out.println(response);

		/*
		 * And we're done!
		 */

		os.close();
		t.close();
	}
}
